/*
 *  PeaGnut - Command-line Gnutella client for Unix
 *  Copyright (C) 2000  Ciaran Lynch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* XXX UNCHECKED XXX */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <netdb.h>

#include "connectionlist.h"
#include "conlist.h"
#include "log.h"
#include "hostfile.h"
#include "packetincoming.h"

#define OUTGOING_CONNECTIONS_MIN 4

int con_list_outgoing_new(void)
{
	/* And now we have a valid set of possible connections */

	if(connection_list_size()<OUTGOING_CONNECTIONS_MIN)
	{
		hostlist_type *new_host;
		int con_socket=-1,outgoing_port=0;
		struct sockaddr_in outgoing_addr;
		struct hostent *hp;

		/* No harm in checking again ... */

		new_host=unlink_next_host();
				
		if(new_host==NULL)
		{
			/*log(LOG_ERR,"Unable to make outgoing connection : Error unlinking next host");*/
			/* Run out of connections ... */
			return 1;
		}
		if((con_socket=socket(PF_INET,SOCK_STREAM,0))==-1)
		{
			log(LOG_ERR,"Unable to make outgoing connection to %s:%s : Error creating outgoing socket : %m",new_host->address,new_host->port);
			free(new_host->address);
			free(new_host->port);
			free(new_host);
			return 1;
		}
		if((fcntl(con_socket,F_SETFL,O_NONBLOCK))==-1)
		{
			log(LOG_ERR,"Unable to make outgoing connection to %s:%s : Error setting socket flags : %m",new_host->address,new_host->port);
			if(close(con_socket)==-1)
				log(LOG_ERR,"Unable to close outgoing socket : %m");
			free(new_host->address);
			free(new_host->port);
			free(new_host);
			return 1;
		}
		
		/* Set outgoing addr */
		if((hp=gethostbyname(new_host->address))==NULL)
		{
			log(LOG_ERR,"Unable to make outgoing connection to %s:%s : Error getting host : %m",new_host->address,new_host->port);
			if(close(con_socket)==-1)
				log(LOG_ERR,"Unable to close outgoing socket : %m");
			free(new_host->address);
			free(new_host->port);
			free(new_host);
			exit(1);
		}
		
		memset(&outgoing_addr, 0, sizeof(outgoing_addr));
		outgoing_addr.sin_family = AF_INET;
		outgoing_addr.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
		sscanf(new_host->port,"%d",&outgoing_port);
		outgoing_addr.sin_port = htons(outgoing_port);
		
		if((connect(con_socket,(struct sockaddr *)&outgoing_addr,sizeof(outgoing_addr)))==-1)
		{
			/* Non-blocking socket won't connect immediately */
			if(errno!=EINPROGRESS)
			{
				log(LOG_ERR,"Unable to make outgoing connection to %s:%s : Error connecting socket : %m",new_host->address,new_host->port);
				if(close(con_socket)==-1)
					log(LOG_ERR,"Unable to close outgoing socket : %m");
				free(new_host->address);
				free(new_host->port);
				free(new_host);
				endhostent();
				return 1;
			}
		}
		if(connection_list_add(new_host->address,new_host->port,CONNECTION_CONNECTING,con_socket))
		{
			log(LOG_ERR,"Unable to make outgoing connection to %s:%s : Error adding outgoing socket to list",new_host->address,new_host->port);
			if(shutdown(con_socket,2)==-1)
				log(LOG_ERR,"Unable to shutdown outgoing socket : %m");
			if(close(con_socket)==-1)
				log(LOG_ERR,"Unable to close outgoing socket : %m");
			free(new_host->address);
			free(new_host->port);
			free(new_host);
			endhostent();
			return 1;
		}
		
		log(LOG_INFO,"Outgoing host added : %s:%s",new_host->address,new_host->port);
		free(new_host->address);
		free(new_host->port);
		free(new_host);
	}
	
	endhostent();
	
	return 0;
}

int con_list_loop(void)
{
	int i,max_fd=0,retval;
	fd_set watch_fds_read,watch_fds_write;
	struct timeval tv;

	FD_ZERO(&watch_fds_read);
	FD_ZERO(&watch_fds_write);
	
	if(connection_list_size()<OUTGOING_CONNECTIONS_MIN)
	{
		con_list_outgoing_new(); /* Handles his own errors */ 
	}
	for(i=0;i<connection_list_size();i++)
	{
		list_type *list_item=connection_list(i);
		if(list_item==NULL)
		{
			log(LOG_ERR,"Unable to process connections : Error returning list item");
			return 1;
		}
		if((list_item->status==CONNECTION_ESTABLISHED)||
		   (list_item->status==CONNECTION_REQUESTED_IN)||
		   (list_item->status==CONNECTION_REQUESTED_OUT)) 
		{
			FD_SET(list_item->con_socket,&watch_fds_read);
		}
		if((list_item->status==CONNECTION_CONNECTING))
		{
			FD_SET(list_item->con_socket,&watch_fds_write);
		}
		if(list_item->con_socket>max_fd) max_fd=list_item->con_socket;

	}

	tv.tv_sec=1;
	tv.tv_usec=0; /* 1 Second */

	max_fd++; /* Set to highest fd + 1 */

	retval=select(max_fd,&watch_fds_read,&watch_fds_write,NULL,&tv);

	if(retval==-1)
	{
		if(errno!=EINTR)
		{
			log(LOG_ERR,"Unable to process connections : Error in select : %m");
		}
	} else if(retval) {
		for(i=0;i<connection_list_size();i++)
		{
			list_type *list_item=connection_list(i);
			if(list_item==NULL)
			{
				log(LOG_ERR,"Unable to process connections : Error returning list item");
				return 1;
			}
			/* The only thing not buffered is connect challenges */
			if(list_item->status==CONNECTION_CONNECTING)
			{
				if(FD_ISSET(list_item->con_socket,&watch_fds_write))
				{
					retval=connection_list_write_challenge(i);
					if(retval==-1) {
						log(LOG_ERR,"Unable to process connection from %s:%s : Error writing to buffer",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else if(retval==0) {
						/* Socket closed */
						log(LOG_ERR,"Unable to process connection from %s:%s : Connection closed",list_item->address,list_item->port);
						connection_list_remove(i); /* Is this right ? */
						i--;
						continue;
					}
				}
			} else if(list_item->status==CONNECTION_REQUESTED_OUT) {
				if(FD_ISSET(list_item->con_socket,&watch_fds_read))
				{
					retval=connection_list_check_challenge(i);
					if(retval==-1) {
						log(LOG_ERR,"Unable to process connection from %s:%s : Error checking challenge",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else if(retval==0) {
						log(LOG_ERR,"Invalid response from %s:%s",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else if(retval==-2) {
						/* Not finished reading */
					}
				}
			} else if(list_item->status==CONNECTION_REQUESTED_IN) {
				if(FD_ISSET(list_item->con_socket,&watch_fds_read))
				{
					retval=connection_list_check_response(i);
					if(retval==-1)
					{
						log(LOG_ERR,"Unable to process connection from %s:%s : Error checking response",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else if(retval==0) {
						log(LOG_ERR,"Invalid response from %s:%s",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else if(retval==1) {
						/* Correct value received ... */
						if(connection_list_write_response(i)==-1)
						{
							log(LOG_ERR,"Unable to write response to %s:%s",list_item->address,list_item->port);
							connection_list_remove(i);
							i--;
						}
						log(LOG_INFO,"Connection established from %s:%s",list_item->address,list_item->port);
					}
				}
			} else if(list_item->status==CONNECTION_ESTABLISHED) {
				/* XXX Remove and do properly */
				outgoing_packet_write_ping(list_item->address,list_item->port);
				if(FD_ISSET(list_item->con_socket,&watch_fds_read))
				{
					/* Read packet to buffer */
					char buffer[4096]; /* Read in 4096 byte chunks */
					retval=read(list_item->con_socket,buffer,sizeof(buffer));
					if(retval==-1)
					{
						if(errno!=EINTR)
						{
							log(LOG_ERR,"Unable to process connection from %s:%s : Error reading from socket : %m",list_item->address,list_item->port);
							connection_list_remove(i);
							i--;
							continue;
						}
					} else if(retval==0) {
						/* Socket closed */
						log(LOG_INFO,"Connection from %s:%s closed",list_item->address,list_item->port);
						connection_list_remove(i);
						i--;
						continue;
					} else {
						/* retval bytes successfully received */
						int processed=0,left=retval,toread;
						while(processed<retval)
						{
							if(list_item->num_packets)
							{
								packet_type *current_packet;
								if(list_item->packet[list_item->num_packets-1].complete==-1)
								{
									/* Add new packet to list */
									packet_type *new_packet=malloc(sizeof(packet_type)*(list_item->num_packets+1));
									memcpy(new_packet,list_item->packet,sizeof(packet_type)*(list_item->num_packets));
									free(list_item->packet);
									list_item->packet=new_packet;
									list_item->packet[list_item->num_packets].complete=0;
									list_item->packet[list_item->num_packets].payload_buffer=NULL;
									(list_item->num_packets)++;
								}
								/* Finish old packet [or new one] */
								current_packet=&(list_item->packet[list_item->num_packets-1]);
								if(current_packet->complete<16)
								{
									/* Read guid */
									if((current_packet->complete+left)>=16) toread=(16-current_packet->complete);
									else toread=left;
									
									memcpy(&(current_packet->guid[current_packet->complete]),&(buffer[processed]),toread);
									processed+=toread;
									left-=toread;
									current_packet->complete+=toread;

								}
								if(left==0) continue;
								if(current_packet->complete<17)
								{
									/* Read type */
									memcpy(&(current_packet->type),&(buffer[processed]),1);
									processed+=1;
									left-=1;
									current_packet->complete+=1;
								}
								if(left==0) continue;
								if(current_packet->complete<18)
								{
									/* Read TTL */
									memcpy(&(current_packet->ttl),&(buffer[processed]),1);
									processed+=1;
									left-=1;
									current_packet->complete+=1;
								}
								if(left==0) continue;
								if(current_packet->complete<19)
								{
									/* Read hops */
									memcpy(&(current_packet->hops),&(buffer[processed]),1);
									processed+=1;
									left-=1;
									current_packet->complete+=1;
								}
								if(left==0) continue;
								if(current_packet->complete<23)
								{
									/* Read payload length */
									if((current_packet->complete+left-19)>=4) toread=(23-current_packet->complete);
									else toread=left;
									
									memcpy(((char *)(&(current_packet->payload_length)))+(current_packet->complete-19),&(buffer[processed]),toread);
									processed+=toread;
									left-=toread;
									current_packet->complete+=toread;
								}
								if(left==0) continue;
								if(current_packet->complete<(23+current_packet->payload_length))
								{
									if(!(current_packet->payload_buffer))
									{
										/* Allocate buffer */
										current_packet->payload_buffer=malloc(current_packet->payload_length);
									}
									if((current_packet->complete+left-23)>=(current_packet->payload_length)) toread=(current_packet->payload_length);/*toread=(23-current_packet->complete+current_packet->payload_length);*/
									else toread=left;

									memcpy(((char *)(current_packet->payload_buffer))+(current_packet->complete-23),&(buffer[processed]),toread);
									processed+=toread;
									left-=toread;
									current_packet->complete+=toread;
								} 
								if(current_packet->complete>=(23+current_packet->payload_length)) {
									/* Packet finished */
									current_packet->complete=-1;
									log(LOG_INFO,"Packet received from %s:%s : Type=%d Payload=%d",list_item->address,list_item->port,current_packet->type,current_packet->payload_length);
								}
							} else {
								/* Create new packet list */
								list_item->packet=malloc(sizeof(packet_type));
								list_item->num_packets=1;
								list_item->packet[0].complete=0;
								list_item->packet[0].payload_buffer=NULL;
							}
						}
					}
				}
			}
		}
	} else {
		/* Nothing ready */
	}
	packet_incoming_process();

	return 0;	
}
