/*
 *  PeaGnut - Command-line Gnutella client for Unix
 *  Copyright (C) 2000  Ciaran Lynch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hostfile.h"
#include "log.h"

/* There's really no point having a linked list of these anymore since they're
 * always accessed from the start anyway.  Instead, they are just stored in an
 * array.  Reading the first element removes it from the list.
 */

hostlist_type *hosts=NULL;
int num_hosts;

int load_hostfile(char *filename)
{
	FILE *host_file;

	if((host_file=fopen(filename,"r"))==NULL)
	{
		log(LOG_ERR,"Unable to open hostfile %s : Error openinf file : %m",filename);
		return 1;
	}
	
	while(!feof(host_file))
	{
		char ip[16],port[6],ch;
		int ip_len=0,port_len=0;
		
		ch=fgetc(host_file);
		while(((ch=='\n')||(ch=='\r'))&&!feof(host_file)) ch=fgetc(host_file);
		while(((ch!=':')&&(ch!='\n'))&&!feof(host_file))
		{
			ip[ip_len++]=ch;
			ch=fgetc(host_file);
		}
		ip[ip_len]=0;
		if(ch==':')
		{
			ch=fgetc(host_file);
			while((ch!='\n')&&!feof(host_file))
			{
				port[port_len++]=ch;
				ch=fgetc(host_file);
			}
			port[port_len]=0;
		} else {
			strcpy(port,"6346");
			while((ch!='\n')&&!feof(host_file)) ch=fgetc(host_file);
		}
		/* ip and port contain values ... */
		if(!feof(host_file))
		{
			if(append_host(ip,port))
			{
				log(LOG_ERR,"Error in hostfile %s : Unable to add host %s:%s",filename,ip,port);
			}
		}
	}
	
	fclose(host_file);

	log(LOG_INFO,"Host file %s successfully loaded",filename);
	
	return 0;
}

int append_host(char *address,char *port)
{	
	hostlist_type *new_hosts;
	int i;
	char *new_address,*new_port;
	
	for(i=0;i<num_hosts;i++)
	{
		if((!strcmp(hosts[i].address,address))&&(!strcmp(hosts[i].port,port))) return 0; /* Duplicate is ok, not an error */
	}
	
	if((new_address=strdup(address))==NULL)
	{
		log(LOG_ERR,"Unable to append host : Out of memory");
		return 1;
	}
	if((new_port=strdup(port))==NULL)
	{
		log(LOG_ERR,"Unable to append host : Out of memory");
		free(new_address);
		return 1;
	}
	
	new_hosts=malloc(sizeof(hostlist_type)*(num_hosts+1));

	if(new_hosts==NULL)
	{
		log(LOG_ERR,"Unable to append host : Out of memory");
		free(new_address);
		free(new_port);
		return 1;
	}
	
	if(num_hosts)
	{
		memcpy(new_hosts,hosts,sizeof(hostlist_type)*num_hosts);
		free(hosts);
	}
	hosts=new_hosts;
	hosts[num_hosts].address=new_address;
	hosts[num_hosts++].port=new_port;

	return 0;
}

int prepend_host(char *address,char *port)
{
	hostlist_type *new_hosts;
	int i;
	char *new_address,*new_port;
	
	for(i=0;i<num_hosts;i++)
	{
		if((!strcmp(hosts[i].address,address))&&(!strcmp(hosts[i].port,port))) return 0; /* Duplicate is ok, not an error */
	}
	
	if((new_address=strdup(address))==NULL)
	{
		log(LOG_ERR,"Unable to prepend host : Out of memory");
		return 1;
	}
	if((new_port=strdup(port))==NULL)
	{
		log(LOG_ERR,"Unable to prepend host : Out of memory");
		free(new_address);
		return 1;
	}
	
	new_hosts=malloc(sizeof(hostlist_type)*(num_hosts+1));
	
	if(new_hosts==NULL)
	{
		log(LOG_ERR,"Unable to prepend host : Out of memory");
		free(new_address);
		free(new_port);
		return 1;
	}
	
	if(num_hosts)
	{
		memcpy(new_hosts+1,hosts,sizeof(hostlist_type)*num_hosts);
		free(hosts);
	}
	hosts=new_hosts;
	hosts[0].address=new_address;
	hosts[0].port=new_port;

	num_hosts++;

	return 0;
}

int free_hosts(void)
{
	int i;
	for(i=0;i<num_hosts;i++)
	{
		free(hosts[i].address);
		free(hosts[i].port);
	}

	if(num_hosts) free(hosts);
	num_hosts=0;

	log(LOG_INFO,"Hosts freed");
	
	return 0;
}

hostlist_type *unlink_next_host(void)
{	
	if(!num_hosts) return NULL;

	if(num_hosts==1)
	{
		num_hosts=0;
		return hosts;
	} else {
		hostlist_type *new_hosts,*return_host;
		
		if((new_hosts=malloc(sizeof(hostlist_type)*(num_hosts-1)))==NULL)
		{
			log(LOG_ERR,"Unable to unlink host : Out of memory");
			return NULL;
		}
				
		if((return_host=malloc(sizeof(hostlist_type)))==NULL)
		{
			log(LOG_ERR,"Unable to unlink host : Out of memory");
			free(new_hosts);
			return NULL;
		}
		memcpy(new_hosts,hosts+1,sizeof(hostlist_type)*(num_hosts-1));
		
		return_host->address=hosts->address;
		return_host->port=hosts->port;

		free(hosts);
		hosts=new_hosts;
		
		num_hosts--;
		
		return return_host;
	}
}
