/*
 *  PeaGnut - Command-line Gnutella client for Unix
 *  Copyright (C) 2000  Ciaran Lynch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>

#include <signal.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h> 

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "log.h"
#include "connection.h"
#include "conlist.h"
#include "hostfile.h"
#include "connectionlist.h"

#include "packetlist.h"


int listen_socket;
int listen_port=6346;
int end_server=0;

int main_server_ip=0,main_server_port=0;

static void terminate_handler(int signum)
{
	end_server=1;
}

static int server_terminated(void)
{
	return end_server;
}

int server_initiate_listen(void)
{
	struct sockaddr_in socket_address;
	char hostnamebuffer[1024]; /* XXX Fixed size buffer shouldn't be problem but ... */
	struct hostent *hp;
	
	signal(SIGTERM,terminate_handler);
	signal(SIGINT,terminate_handler);
	
	signal(SIGHUP,SIG_IGN);
	signal(SIGPIPE,SIG_IGN);
	
	siginterrupt(SIGINT,0);
	siginterrupt(SIGTERM,0);
	
	main_server_ip=0;
	main_server_port=0;
	
	if((listen_socket=socket(AF_INET,SOCK_STREAM,0))==-1)
	{
		log(LOG_ERR,"Error creating main listen socket : %m");
		exit(1); /* OK */
	}
	if((fcntl(listen_socket,F_SETFL,O_NONBLOCK))==-1)
	{
		log(LOG_ERR,"Unable to set main listen socket flags : %m");
		if(close(listen_socket)==-1)
			log(LOG_ERR,"Unable to close main listen socket : %m");
		exit(1); /* OK */
	}
	
	memset((void *)&socket_address,0,sizeof(socket_address));

	socket_address.sin_family=AF_INET;
	socket_address.sin_addr.s_addr=INADDR_ANY;
	socket_address.sin_port=htons(listen_port);

	if(bind(listen_socket,(struct sockaddr *)&socket_address,sizeof(socket_address))==-1)
	{
		log(LOG_ERR,"Error binding main listen socket : %m");
		if(close(listen_socket)==-1)
			log(LOG_ERR,"Unable to close main listen socket : %m");
		exit(1); /* OK */
	}

	if(listen(listen_socket,20)==-1)
	{
		log(LOG_ERR,"Error listening to socket : %m");
		if(close(listen_socket)==-1)
			log(LOG_ERR,"Unable to close main listen socket : %m");
		exit(1); /* OK */
	}
	
	/* Is there an easier way to do this ? */
	gethostname(hostnamebuffer,1024);
	hp=gethostbyname(hostnamebuffer);
	endhostent();
	
	main_server_ip=ntohl((((struct in_addr *)(hp->h_addr)))->s_addr);
	main_server_port=ntohs(socket_address.sin_port);

	log(LOG_NOTICE,"Server started on %s:%d",inet_ntoa(*(struct in_addr *)(hp->h_addr)),listen_port);
	
	return listen_socket;
}

int server_main_loop(void)
{
	int socket_status=0,i;
	fd_set read_socketset;
	struct timeval tv;

	load_hostfile("hostlist"); /* XXX Needs more flexibility*/
	
	while(!server_terminated())
	{
		FD_ZERO(&read_socketset);
		FD_SET(listen_socket,&read_socketset);

		tv.tv_sec=1;/*0;*/
		tv.tv_usec=10000; /* 10 ms */

		socket_status=select(listen_socket+1,&read_socketset,NULL,NULL,&tv);
		if(socket_status==-1)
		{
			if(errno!=EINTR)
			{
				log(LOG_ERR,"Error in main listen socket select : %m");
				if(close(listen_socket)==-1)
					log(LOG_ERR,"Unable to close main listen socket : %m");
				exit(1); /* OK */
			}
		} else if (socket_status>0) {
			/* Stuff ready */
			connection_handle(listen_socket);
			/* Ignore errors since we can't really do anything
		   	 * about them
	      		 */		 
		} 
	
		if(con_list_loop())
		{
			log(LOG_ERR,"Error in connection main loop");
		}
	}
	
	log(LOG_NOTICE,"Server finished on port %d",listen_port);
	
	for(i=0;i<connection_list_size();i++)
	{
		list_type *list_item=connection_list(i);
		if(shutdown(list_item->con_socket,2)==-1)
			log(LOG_ERR,"Unable to shutdown socket : %m");
		if(close(list_item->con_socket)==-1)
			log(LOG_ERR,"Unable to close socket : %m");
	}
	
	free_hosts();
	connection_list_free();
	packet_list_free();
	
	return 0;
}

int server_close_listen(void)
{
	if(close(listen_socket)==-1)
	{
		log(LOG_ERR,"Unable to close main listen socket : %m");
		return 1;
	}
	
	return 0;
}
